##########################
### required libraries ###
##########################
library(rseg)
library(partykit)

#########################
### Application study ###
#########################

# Load data from the respective labraries
library(evtree)
data("StatlogHeart")
StatlogHeart$heart_disease <- as.factor(as.numeric(StatlogHeart$heart_disease) - 1)
data("GermanCredit")
GermanCredit$credit_risk <- as.factor(as.numeric(GermanCredit$credit_risk) - 1)
data("ContraceptiveChoice")
ContraceptiveChoice$contraceptive_method_used <- as.factor(as.numeric(ContraceptiveChoice$contraceptive_method_used != "no use"))

library(mlbench)
data(BostonHousing)
data(BreastCancer)
BreastCancer$Class <- as.factor(as.numeric(BreastCancer$Class) - 1)
BreastCancer <- subset(BreastCancer, select = -Id)
for (i in 1:9) BreastCancer[, i] <- as.numeric(BreastCancer[, i])
data(Ionosphere)
Ionosphere$Class <- as.factor(as.numeric(Ionosphere$Class) - 1)
Ionosphere <- subset(Ionosphere, select = -V2)
data(Ozone)
Ozone$V1 <- as.numeric(as.character(Ozone$V1))
Ozone$V2 <- as.numeric(as.character(Ozone$V2))
Ozone$V3 <- as.numeric(as.character(Ozone$V3))
data(PimaIndiansDiabetes2)
PimaIndiansDiabetes2$diabetes <- as.factor(as.numeric(PimaIndiansDiabetes2$diabetes) - 1)
data(Servo)
data(Sonar)
Sonar$Class <- as.factor(as.numeric(Sonar$Class) - 1)

library(kernlab)
data(musk)
data(promotergene)
promotergene$Class <- as.factor(as.numeric(promotergene$Class) - 1)
data(spam)
spam$type <- as.factor(as.numeric(spam$type) - 1)

library(TH.data)
data("birds")
data("bodyfat")
data("GlaucomaM")
GlaucomaM$Class <- as.factor(as.numeric(GlaucomaM$Class) - 1)

library(catdata)
data(aids)
aids <- subset(aids, select = -person)
aids$drugs <- as.factor(aids$drugs)

library(openintro)
data(gifted)
data(mammals)
mammals <- subset(mammals, select = -Species)

library(MASS)
data(birthwt)
birthwt <- subset(birthwt, select = -low)
birthwt$race <- as.factor(birthwt$race)
birthwt$smoke <- as.factor(birthwt$smoke)
birthwt$ht <- as.factor(birthwt$ht)
birthwt$ui <- as.factor(birthwt$ui)
data(Insurance)
data(quine)
data(shuttle)
shuttle$use <- as.factor(as.numeric(shuttle$use) - 1)
data(UScrime)
UScrime$So <- as.factor(UScrime$So)

library(ElemStatLearn)
data(prostate)
prostate <- subset(prostate, select = -train)
prostate$svi <- as.factor(prostate$svi)

data(SAheart)
SAheart$chd <- as.factor(as.numeric(SAheart$chd))

# vectors of data names to devide classification and regression problems
datasets.class <- c("StatlogHeart", "GermanCredit", "ContraceptiveChoice", "BreastCancer", "Ionosphere", "PimaIndiansDiabetes2", "Sonar", "musk", "promotergene", "spam", "GlaucomaM", "shuttle", "SAheart")
datasets.reg  <- c("BostonHousing", "Ozone", "Servo", "birds", "bodyfat", "aids", "gifted", "mammals", "birthwt", "Insurance", "quine", "UScrime", "prostate")

# vectors of formulas corresponding to the problems
form.class <- list(heart_disease ~ ., credit_risk ~ ., contraceptive_method_used ~ ., Class ~ ., Class ~ ., diabetes ~ ., Class ~ ., 
                  Class ~ ., Class ~ ., type ~ ., Class ~ ., use ~ ., chd ~ .)
form.reg  <- list(medv ~ ., V4 ~ ., Class ~ ., SG5 ~ GST + DBH + AOT + AFS + DWC + LOG, DEXfat ~ ., cd4 ~ ., score ~ ., BrainWt ~ ., 
                  bwt ~ ., Claims ~ ., Days ~ ., y ~ ., lpsa ~ .)

###########################################
### Analysis of classification problems ###
###########################################
niter <- 250 # number of bootstrap samples
results.class <- list()

for (i in 1:length(datasets.class)) { # slope over data sets
  evaldat <- na.omit(get(datasets.class[i])) # only complete observations as evtree doesn't support missing values to date
  n <- nrow(evaldat)
  p <- ncol(evaldat)
  minsplt <- max(20, ceiling(n * 0.09)) # there need to be at least 9% observations in a node to allow further splitting
  minbukt <- max(7, ceiling(n * 0.03)) # a terminal node needs to contain at least 3% of the observations
  set.seed(1234) # for reproducability
  ctreeBrier <- etreeBrier <- rtreeBrier <- cBrier <- rBrier <- eBrier <- eBrier.aic <- eBrier.bic <- c() # vectors to contain error estimates
  ctreeN <- etreeN <- rtreeN <- cN <- rN <- eN <- eN.aic <- eN.bic <- c() # vectors to contain the complexity of models
  
  for (j in 1:niter) { # slope over bootstrap samples
    cat("\r", "Dataset processed: ", i, " (", datasets.class[i], ", n = ", n, ", p = ", p - 1, "); iteration: ", j, "                          ", sep = ""); flush.console()
    samp <- sample(1:nrow(evaldat), nrow(evaldat), replace = TRUE) # bootstrap sample of the data
    
    # Recursive segmentation
    cmod <- cSeg(form.class[[i]], evaldat[samp, ], minsplit = minsplt, minbucket = minbukt) # conditional inference
    rmod <- rSeg(form.class[[i]], evaldat[samp, ], minsplit = minsplt, minbucket = minbukt) # CART
    emod <- eSeg(form.class[[i]], evaldat[samp, ], minsplit = minsplt, minbucket = minbukt) # evolutionary learning
    emod.aic <- aic(emod, family = "binomial") # evolutionary learning by AIC
    emod.bic <- bic(emod, family = "binomial") # evolutionary learning by BIC
    
    # Recursive partitioning (trees)
    ctreemod <- gettree(cmod) # conditional inference
    rtreemod <- gettree(rmod) # CART
    etreemod <- gettree(emod) # evolutionary learning
    
    # Estimation of Brier Score
    ctreeBrier <- c(ctreeBrier, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(ctreemod, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    rtreeBrier <- c(rtreeBrier, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(rtreemod, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    etreeBrier <- c(etreeBrier, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(etreemod, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    
    cBrier <- c(cBrier, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(cmod, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    rBrier <- c(rBrier, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(rmod, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    eBrier <- c(eBrier, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(emod, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    eBrier.aic <- c(eBrier.aic, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(emod.aic, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    eBrier.bic <- c(eBrier.bic, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.class[[i]])[1]])) - predict(emod.bic, newdata = evaldat[-samp, ], type = "prob")[, "1"])^2))
    
    # Complexity of models
    ctreeN <- c(ctreeN, width(ctreemod))
    rtreeN <- c(rtreeN, width(rtreemod))
    etreeN <- c(etreeN, width(etreemod))
    
    cN <- c(cN, length(cmod))
    rN <- c(rN, length(rmod))
    eN <- c(eN, length(emod))
    eN.aic <- c(eN.aic, length(emod.aic))
    eN.bic <- c(eN.bic, length(emod.bic))
  }
  
  Brier <- rbind(ctreeBrier, etreeBrier, rtreeBrier, cBrier, rBrier, eBrier, eBrier.aic, eBrier.bic)
  N <- rbind(ctreeN, etreeN, rtreeN, cN, rN, eN, eN.aic, eN.bic)
  rownames(Brier) <- rownames(N) <- c("ctree", "evtree", "rpart", "cseg", "rseg", "eseg", "eseg.aic", "eseg.bic")
  colnames(Brier) <- colnames(N) <- 1:niter
  result <- list(Brier, N)
  names(result) <- c("Brier", "N")
  results.class[[i]] <- result
}
names(results.class) <- datasets.class

#######################################
### Analysis of regression problems ###
#######################################
niter <- 250 # number of bootstrap samples
results.reg <- list()

for (i in 1:length(datasets.reg)) { # slope over data sets
  evaldat <- na.omit(get(datasets.reg[i])) # only complete observations as evtree doesn't support missing values to date
  n <- nrow(evaldat)
  p <- ncol(evaldat)
  minsplt <- max(20, ceiling(n * 0.09)) # there need to be at least 9% observations in a node to allow further splitting
  minbukt <- max(7, ceiling(n * 0.03)) # a terminal node needs to contain at least 3% of the observations
  set.seed(1234) # for reproducability
  ctreeMSE <- etreeMSE <- rtreeMSE <- cMSE <- rMSE <- eMSE <- eMSE.aic <- eMSE.bic <- c() # vectors to contain error estimates
  ctreeN <- etreeN <- rtreeN <- cN <- rN <- eN <- eN.aic <- eN.bic <- c() # vectors to contain the complexity of models
  
  for (j in 1:niter) { # slope over bootstrap samples
    cat("\r", "Dataset processed: ", i, " (", datasets.reg[i], ", n = ", n, ", p = ", p - 1, "); iteration: ", j, "                          ", sep = ""); flush.console()
    samp <- sample(1:nrow(evaldat), nrow(evaldat), replace = TRUE) # bootstrap sample of the data
    
    # Recursive segmentation
    cmod <- cSeg(form.reg[[i]], evaldat[samp, ], minsplit = minsplt, minbucket = minbukt) # conditional inference
    rmod <- rSeg(form.reg[[i]], evaldat[samp, ], minsplit = minsplt, minbucket = minbukt) # CART
    emod <- eSeg(form.reg[[i]], evaldat[samp, ], minsplit = minsplt, minbucket = minbukt) # evolutionary learning
    emod.aic <- aic(emod) # evolutionary learning by AIC
    emod.bic <- bic(emod) # evolutionary learning by BIC
    
    # Recursive partitioning (trees)
    ctreemod <- gettree(cmod) # conditional inference
    rtreemod <- gettree(rmod) # CART
    etreemod <- gettree(emod) # evolutionary learning
    
    # Estimation of MSE
    ctreeMSE <- c(ctreeMSE, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(ctreemod, newdata = evaldat[-samp, ]))^2))
    rtreeMSE <- c(rtreeMSE, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(rtreemod, newdata = evaldat[-samp, ]))^2))
    etreeMSE <- c(etreeMSE, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(etreemod, newdata = evaldat[-samp, ]))^2))
    
    cMSE <- c(cMSE, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(cmod, newdata = evaldat[-samp, ]))^2))
    rMSE <- c(rMSE, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(rmod, newdata = evaldat[-samp, ]))^2))
    eMSE <- c(eMSE, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(emod, newdata = evaldat[-samp, ]))^2))
    eMSE.aic <- c(eMSE.aic, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(emod.aic, newdata = evaldat[-samp, ]))^2))
    eMSE.bic <- c(eMSE.bic, mean((as.numeric(as.character(evaldat[-samp, all.vars(form.reg[[i]])[1]])) - predict(emod.bic, newdata = evaldat[-samp, ]))^2))
    
    # Complexity of models
    ctreeN <- c(ctreeN, width(ctreemod))
    rtreeN <- c(rtreeN, width(rtreemod))
    etreeN <- c(etreeN, width(etreemod))
    
    cN <- c(cN, length(cmod))
    rN <- c(rN, length(rmod))
    eN <- c(eN, length(emod))
    eN.aic <- c(eN.aic, length(emod.aic))
    eN.bic <- c(eN.bic, length(emod.bic))
  }
  
  MSE <- rbind(ctreeMSE, etreeMSE, rtreeMSE, cMSE, rMSE, eMSE, eMSE.aic, eMSE.bic)
  N <- rbind(ctreeN, etreeN, rtreeN, cN, rN, eN, eN.aic, eN.bic)
  rownames(MSE) <- rownames(N) <- c("ctree", "evtree", "rpart", "cseg", "rseg", "eseg", "eseg.aic", "eseg.bic")
  colnames(MSE) <- colnames(N) <- 1:niter
  result <- list(MSE, N)
  names(result) <- c("MSE", "N")
  results.reg[[i]] <- result
}
names(results.reg) <- datasets.reg

#######################
### Grafical Output ###
#######################

# classification problem
error.ratio.c <- sapply(results.class, function(x) x[["Brier"]]["ctree", ] / x[["Brier"]]["cseg", ])
error.ratio.e <- sapply(results.class, function(x) x[["Brier"]]["evtree", ] / x[["Brier"]]["eseg.bic", ])
error.ratio.r <- sapply(results.class, function(x) x[["Brier"]]["rpart", ] / x[["Brier"]]["rseg", ])

complex.ratio.c <- sapply(results.class, function(x) x[["N"]]["ctree", ] / x[["N"]]["cseg", ])
complex.ratio.e <- sapply(results.class, function(x) x[["N"]]["evtree", ] / x[["N"]]["eseg.bic", ])
complex.ratio.r <- sapply(results.class, function(x) x[["N"]]["rpart", ] / x[["N"]]["rseg", ])

yrange.error <- range(error.ratio.c, error.ratio.e, error.ratio.r)
yrange.complex <- range(complex.ratio.c, complex.ratio.e, complex.ratio.r)
  
par(mfrow = c(4, 2), mar = c(3, 0, 2, 1) + 0.1, omi = c(0.5, 2, 0, 0))

boxplot(error.ratio.c, horizontal = TRUE, las = 1, ylim = yrange.error, log = "x")
abline(v=1)
mtext("conditional inference", side = 3, line = 0.5, adj = 0)
boxplot(complex.ratio.c, horizontal = TRUE, las = 1, yaxt = "n", ylim = yrange.complex, log = "x")
abline(v=1)

boxplot(error.ratio.e, horizontal = TRUE, las = 1, ylim = yrange.error, log = "x")
abline(v=1)
mtext("evolutionary learning", side = 3, line = 0.5, adj = 0)
boxplot(complex.ratio.e, horizontal = TRUE, las = 1, yaxt = "n", ylim = yrange.complex, log = "x")
abline(v=1)

boxplot(error.ratio.r, horizontal = TRUE, las = 1, ylim = yrange.error, log = "x")
abline(v=1)
mtext("CART", side = 3, line = 0.5, adj = 0)
boxplot(complex.ratio.r, horizontal = TRUE, las = 1, yaxt = "n", ylim = yrange.complex, log = "x")
abline(v=1)

library(beeswarm)
boxplot(list(apply(error.ratio.r, 2, mean), apply(error.ratio.e, 2, mean), apply(error.ratio.c, 2, mean)), horizontal = TRUE, names = c("CART", "evo.learn.", "cond.inf."), las = 1)
beeswarm(list(apply(error.ratio.r, 2, mean), apply(error.ratio.e, 2, mean), apply(error.ratio.c, 2, mean)), horiz = TRUE, add = TRUE, pch = 16)
abline(v=1)
mtext("error (part. / seg.)", side = 1, line = 4)
boxplot(list(apply(complex.ratio.r, 2, mean), apply(complex.ratio.e, 2, mean), apply(complex.ratio.c, 2, mean)), horizontal = TRUE, yaxt = "n")
beeswarm(list(apply(complex.ratio.r, 2, mean), apply(complex.ratio.e, 2, mean), apply(complex.ratio.c, 2, mean)), horiz = TRUE, add = TRUE, pch = 16)
abline(v=1)
mtext("complexity (part. / seg.)", side = 1, line = 4)

# regression problem
error.ratio.c <- sapply(results.reg, function(x) x[["MSE"]]["ctree", ] / x[["MSE"]]["cseg", ])
error.ratio.e <- sapply(results.reg, function(x) x[["MSE"]]["evtree", ] / x[["MSE"]]["eseg.bic", ])
error.ratio.r <- sapply(results.reg, function(x) x[["MSE"]]["rpart", ] / x[["MSE"]]["rseg", ])

complex.ratio.c <- sapply(results.reg, function(x) x[["N"]]["ctree", ] / x[["N"]]["cseg", ])
complex.ratio.e <- sapply(results.reg, function(x) x[["N"]]["evtree", ] / x[["N"]]["eseg.bic", ])
complex.ratio.r <- sapply(results.reg, function(x) x[["N"]]["rpart", ] / x[["N"]]["rseg", ])

yrange.error <- range(error.ratio.c, error.ratio.e, error.ratio.r)
yrange.complex <- range(complex.ratio.c, complex.ratio.e, complex.ratio.r)

par(mfrow = c(4, 2), mar = c(3, 0, 2, 1) + 0.1, omi = c(0.5, 2, 0, 0))

boxplot(error.ratio.c, horizontal = TRUE, las = 1, ylim = yrange.error, log = "x")
abline(v=1)
mtext("conditional inference", side = 3, line = 0.5, adj = 0)
boxplot(complex.ratio.c, horizontal = TRUE, las = 1, yaxt = "n", ylim = yrange.complex, log = "x")
abline(v=1)

boxplot(error.ratio.e, horizontal = TRUE, las = 1, ylim = yrange.error, log = "x")
abline(v=1)
mtext("evolutionary learning", side = 3, line = 0.5, adj = 0)
boxplot(complex.ratio.e, horizontal = TRUE, las = 1, yaxt = "n", ylim = yrange.complex, log = "x")
abline(v=1)

boxplot(error.ratio.r, horizontal = TRUE, las = 1, ylim = yrange.error, log = "x")
abline(v=1)
mtext("CART", side = 3, line = 0.5, adj = 0)
boxplot(complex.ratio.r, horizontal = TRUE, las = 1, yaxt = "n", ylim = yrange.complex, log = "x")
abline(v=1)

library(beeswarm)
boxplot(list(apply(error.ratio.r, 2, mean), apply(error.ratio.e, 2, mean), apply(error.ratio.c, 2, mean)), horizontal = TRUE, names = c("CART", "evo.learn.", "cond.inf."), las = 1)
beeswarm(list(apply(error.ratio.r, 2, mean), apply(error.ratio.e, 2, mean), apply(error.ratio.c, 2, mean)), horiz = TRUE, add = TRUE, pch = 16)
abline(v=1)
mtext("error (part. / seg.)", side = 1, line = 4)
boxplot(list(apply(complex.ratio.r, 2, mean), apply(complex.ratio.e, 2, mean), apply(complex.ratio.c, 2, mean)), horizontal = TRUE, yaxt = "n")
beeswarm(list(apply(complex.ratio.r, 2, mean), apply(complex.ratio.e, 2, mean), apply(complex.ratio.c, 2, mean)), horiz = TRUE, add = TRUE, pch = 16)
abline(v=1)
mtext("complexity (part. / seg.)", side = 1, line = 4)








