##########################
### required libraries ###
##########################
library(rseg)
library(partykit)

############################################################
### function to compute the normlized mutual information ###
############################################################
NMI <- function(x, y) {
  tab <- table(x, y)
  n <- length(x)
  bval <- nrow(tab)*ncol(tab)
  res <- c()
  for (i in rownames(tab)) {
    for (j in colnames(tab)) {
      res <- c(res, tab[i,j] * log((tab[i,j]*n)/(sum(tab[i,])*sum(tab[,j])), base = bval))
    }
  }
  res <- (2/n) * sum(res, na.rm = TRUE)
  res
}

########################
### Simulation study ###
########################
deltas <- seq(0, 3, length.out = 10) # initiate sequence of delta values = effect

reps <- 500 # number of simulation runs
resultlist <- lapply(1:9, function(x) array(NA, dim = c(reps, length(deltas), 2))) # list of arrays to take results

set.seed(1234) # for reproducability
for (j in 1:length(deltas)) { # slope over each realization of delta
  delta <- deltas[j]
  for (i in 1:reps) { # slope over simulation runs
    cat("\r", "effect size: ", delta, ", simulation run: ", i, "              ", sep = "") # display process of computations
    
    mydat <- as.data.frame(sapply(1:10, function(x) round(runif(1000, 0, 1), 3))) # initiate data set
    seg <- with(mydat, I(V1 >= 0.5 & V1 <= 0.8 & V2 >= 0.5 & V2 <= 0.8) * 1 + I(V1 >= 0.1 & V1 <= 0.3 & V2 >= 0.1 & V2 <= 0.3) * 2) # define segments
    mue <- c(0, delta, -delta)[seg + 1] # expectation mue
    mydat$Y <- rnorm(1000, mue, 1) # generate response
    
    # Conditional inference
    cseg <- cSeg(Y ~ ., data = mydat) # recursive segmentation
    citree <- gettree(cseg) # tree
    
    # Evolutionary learning
    eseg <- eSeg(Y ~ ., data = mydat) # recursive segmentation
    eseg.aic <- aic(eseg)  # recursive segmentation, first segments chosen by AIC criterion
    eseg.bic <- bic(eseg)  # recursive segmentation, first segments chosen by BIC criterion
    etree <- gettree(eseg) # tree
    
    # CART
    rseg <- rSeg(Y ~ ., data = mydat) # recursive segmentation
    cartree <- gettree(rseg) # tree
    
    # PRIM
    prim <- PRIM_bs(Y ~ ., data = mydat, beta_min = 7/1000, f_min = 2, peel_alpha = c(0.05, 0.1), target = mean, alter_crit = T, paste_alpha = 0.01, max_steps = 20, stop_by_dec = FALSE, print_position = FALSE, B = 5+1) # PRIM using bootstrapped samples and multiple alhpa values
    
    # Computation of the ANMI for each model
    ref <- NMI(seg, seg) # maximally possible NMI
    resultlist[[1]][i, j, 1] <- NMI(predict(citree, type = "node"), seg)/ref
    resultlist[[2]][i, j, 1] <- NMI(predict(cseg, type = "segment"), seg)/ref
    resultlist[[3]][i, j, 1] <- NMI(predict(etree, type = "node"), seg)/ref
    resultlist[[4]][i, j, 1] <- NMI(predict(eseg, type = "segment"), seg)/ref
    resultlist[[5]][i, j, 1] <- NMI(predict(eseg.aic, type = "segment"), seg)/ref
    resultlist[[6]][i, j, 1] <- NMI(predict(eseg.bic, type = "segment"), seg)/ref 
    resultlist[[7]][i, j, 1] <- NMI(if(class(cartree)[1] == "partynode") rep(1, n) else predict(cartree, type = "node"), seg)/ref
    resultlist[[8]][i, j, 1] <- NMI(predict(rseg, type = "segment"), seg)/ref
    resultlist[[9]][i, j, 1] <- NMI(predict(prim, type = "box"), seg)/ref
    
    # Assessment of the number of detected subgroups by each model
    resultlist[[1]][i, j, 2] <- width(citree)
    resultlist[[2]][i, j, 2] <- length(cseg)
    resultlist[[3]][i, j, 2] <- width(etree)
    resultlist[[4]][i, j, 2] <- length(eseg)
    resultlist[[5]][i, j, 2] <- length(eseg.aic)
    resultlist[[6]][i, j, 2] <- length(eseg.bic)
    resultlist[[7]][i, j, 2] <- width(cartree)
    resultlist[[8]][i, j, 2] <- length(rseg)
    resultlist[[9]][i, j, 2] <- length(prim$f)
  }
}
cat("\n")

#######################
### Grafical Output ###
#######################

###########################
### adjusted NMI (ANMI) ###
###########################
par(mfrow = c(1, 4), mar = c(0, 0, 0, 0), oma = c(5, 5, 4, 1))

for (i in c(1,2,3,6,7,8,9)) {
  if (i %in% c(1,3,7,9)) {
    plot(1, 1, xlim = c(0,3), ylim = c(0,1), type = "n", xlab = expression(delta), ylab = ifelse(i == 1, "ANMI", ""), las = 1, xpd = NA, cex.lab = 2, cex.axis = 1.5, xaxt = "n", yaxt = ifelse(i == 1, "s", "n"))
    axis(1, at = 0:3, labels = 0:3, cex.axis = 1.5)
  }
  mtext(ifelse(i %in% c(1,2), "ctree", ifelse(i %in% c(3,6), "evtree", ifelse(i %in% c(7,8), "CART", "PRIM"))), side = 3, line = 1, at = 1.5, cex = 1.5)
  polygon(c(deltas, rev(deltas)), c(apply(resultlist[[i]][, , 1], 2, function(x) quantile(x, 0.25)), rev(apply(resultlist[[i]][, , 1], 2, function(x) quantile(x, 0.75)))), col = adjustcolor("grey", alpha.f = 0.5), border = FALSE)
  lines(deltas, apply(resultlist[[i]][, , 1], 2, median), lty = ifelse(i %in% c(1, 3, 7, 9), 1, 2), lwd = 2)
  if (i == 1) legend("topleft", c("rpart", "rseg"), lty = 1:2, lwd = 2, bty = "n", cex = 2)
}

###############+#############################
### table of number of detected subgroups ###
#############################################
maxseg <- max(resultlist[[1]][, , 2], resultlist[[2]][, , 2], resultlist[[3]][, , 2], resultlist[[5]][, , 2], resultlist[[7]][, , 2], resultlist[[8]][, , 2], resultlist[[9]][, , 2])

resultmat <- rbind(
  t(sapply(1:10, function(x) table(resultlist[[1]][, x, 2])[as.character(1:maxseg)])),
  t(sapply(1:10, function(x) table(resultlist[[2]][, x, 2])[as.character(1:maxseg)])),
  t(sapply(1:10, function(x) table(resultlist[[3]][, x, 2])[as.character(1:maxseg)])),
  t(sapply(1:10, function(x) table(resultlist[[6]][, x, 2])[as.character(1:maxseg)])),
  t(sapply(1:10, function(x) table(resultlist[[7]][, x, 2])[as.character(1:maxseg)])),
  t(sapply(1:10, function(x) table(resultlist[[8]][, x, 2])[as.character(1:maxseg)])),
  t(sapply(1:10, function(x) table(resultlist[[9]][, x, 2])[as.character(1:maxseg)]))
)
colnames(resultmat) <- 1:ncol(resultmat)
resultmat <- cbind(rep(round(deltas, 2), 7), resultmat)

library(xtable)
xtable(resultmat)  
